ETH_RESET_CMD = b"\x12\x00\x00\x00\x00"
ETH_ANSWER_ERR = 0x22
ETH_ANSWER_NOQUEUE = 0x23
DP_MODULE_MAX_COEFF = 16
NUM_MAX_COEFFS_DSP = 12
NUM_DSP_CLASSES = 7

ETH_CMD_REQUEST = b"\x11"
COM_FUNCTION = "\x50"
COM_WRITE_VAR = "\x20"
WRITE_FLOAT_SIZE_PAYLOAD = "\x00\x05"
WRITE_DOUBLE_SIZE_PAYLOAD = "\x00\x03"
COM_READ_VAR = "\x10\x00\x01"
COM_REQUEST_CURVE = "\x40"
COM_SEND_WFM_REF = "\x41"

UDC_FIRMWARE_VERSION = "0.44 2022-06-30"

ufm_offset = {
    "serial": 0,
    "calibdate": 4,
    "variant": 9,
    "rburden": 10,
    "calibtemp": 12,
    "vin_gain": 14,
    "vin_offset": 16,
    "iin_gain": 18,
    "iin_offset": 20,
    "vref_p": 22,
    "vref_n": 24,
    "gnd": 26,
}

type_format = {
    "uint8_t": "BBHBB",
    "uint16_t": "BBHHB",
    "uint32_t": "BBHIB",
    "float": "BBHfB",
}

bytes_format = {"Uint16": "H", "Uint32": "L", "Uint64": "Q", "float": "f"}

type_size = {"uint8_t": 6, "uint16_t": 7, "uint32_t": 9, "float": 9}

num_blocks_curves_fax = [16, 16, 16]
size_curve_block = [1024, 1024, 1024]

ListVar = [
    "iLoad1",
    "iLoad2",
    "iMod1",
    "iMod2",
    "iMod3",
    "iMod4",
    "vLoad",
    "vDCMod1",
    "vDCMod2",
    "vDCMod3",
    "vDCMod4",
    "vOutMod1",
    "vOutMod2",
    "vOutMod3",
    "vOutMod4",
    "temp1",
    "temp2",
    "temp3",
    "temp4",
    "ps_OnOff",
    "ps_OpMode",
    "ps_Remote",
    "ps_OpenLoop",
    "ps_SoftInterlocks",
    "ps_HardInterlocks",
    "iRef",
    "wfmRef_Gain",
    "wfmRef_Offset",
    "sigGen_Enable",
    "sigGen_Type",
    "sigGen_Ncycles",
    "sigGenPhaseStart",
    "sigGen_PhaseEnd",
    "sigGen_Freq",
    "sigGen_Amplitude",
    "sigGen_Offset",
    "sigGen_Aux",
    "dp_ID",
    "dp_Class",
    "dp_Coeffs",
    "ps_Model",
    "wfmRef_PtrBufferStart",
    "wfmRef_PtrBufferEnd",
    "wfmRef_PtrBufferK",
    "wfmRef_SyncMode",
]


hradc_variant = [
    "HRADC-FBP",
    "HRADC-FAX-A",
    "HRADC-FAX-B",
    "HRADC-FAX-C",
    "HRADC-FAX-D",
]

hradc_input_types = [
    "GND",
    "Vref_bipolar_p",
    "Vref_bipolar_n",
    "Temp",
    "Vin_bipolar_p",
    "Vin_bipolar_n",
    "Iin_bipolar_p",
    "Iin_bipolar_n",
]

num_dsp_modules = [4, 4, 4, 6, 8, 4, 2, 2]
num_coeffs_dsp_modules = [0, 1, 1, 4, 8, 16, 2]
dsp_classes_names = [
    "DSP_Error",
    "DSP_SRLim",
    "DSP_LPF",
    "DSP_PI",
    "DSP_IIR_2P2Z",
    "DSP_IIR_3P3Z",
    "DSP_VdcLink_FeedForward",
    "DSP_Vect_Product",
]

num_blocks_curves_fbp = [4, 4, 4]



import re
import socket
import struct
import typing

import serial

from .base import BaseDRS
from .consts import ETH_ANSWER_ERR, ETH_CMD_REQUEST, ETH_RESET_CMD
from .utils import checksum, get_logger
from .validation import SerialErrPckgLen, validate

logger = get_logger(name=__file__)


class SerialDRS(BaseDRS):
    """DRS communication through serial ports"""

    def __init__(self, port: str = None, baud: int = 115200):
        super().__init__()
        self.ser: typing.Optional[serial.Serial] = None

        if port is None:
            print(
                "From 2.0.0 onwards, creating the object then using 'connect' to connect will be deprecated. Please use 'SerialDRS(port, baud)' instead."
            )
        else:
            self.connect(port, baud)

    def _transfer_write(self, msg: str):
        full_msg = (self._slave_addr + msg).encode("ISO-8859-1")
        self.ser.write(checksum(full_msg))

    @validate
    def _transfer(self, msg: str, size: int) -> bytes:
        self._transfer_write(msg)
        return self.ser.read(size)

    def _reset_input_buffer(self):
        self.ser.reset_input_buffer()

    @property
    def timeout(self) -> float:
        return self.ser.timeout

    @timeout.setter
    def timeout(self, new_timeout):
        self.ser.timeout = new_timeout

    def is_open(self) -> bool:
        return self.ser.isOpen()

    def connect(self, port: str = "COM2", baud: int = 115200):
        if self.ser and self.ser.is_open:
            logger.warning(
                "PyDRS obj {} serial port {} is already open. Disconnect before opening a new connection.".format(
                    self, self.ser
                )
            )
            return False
        try:
            self.ser = serial.Serial(
                port, baud, timeout=1
            )  # port format should be 'COM'+number
            return True
        except Exception:
            logger.exception("Failed to open serial port ({}, {})".format(port, baud))
            return False

    def disconnect(self):
        if not self.ser or not self.ser.is_open:
            return True

        try:
            self.ser.close()
            return True
        except Exception:
            logger.exception("Failed to disconnect serial port ({})".format(self.ser))
            return False


class EthDRS(BaseDRS):
    """DRS communication through TCP/IP"""

    def __init__(self, address: str = None, port: int = 5000):
        super().__init__()
        self._serial_timeout = 50
        if address is None:
            print(
                "From 2.0.0 onwards, creating the object then using 'connect' to connect will be deprecated. Please use 'EthDRS(address, port)' instead."
            )
        else:
            self.connect(address, port)

    def _format_message(self, msg: bytes, msg_type: bytes) -> bytes:
        msg = msg_type + struct.Struct(">f").pack(self._serial_timeout) + msg
        return msg[0:1] + struct.pack(">I", (len(msg) - 1)) + msg[1:]

    def _reset_input_buffer(self):
        self.socket.sendall(ETH_RESET_CMD)
        self.socket.recv(16)

    @staticmethod
    def _parse_reply_size(reply: bytes) -> int:
        return struct.unpack(">I", reply[1:])[0]

    def _get_reply(self, size: int = None) -> bytes:
        data_size = self._parse_reply_size(self.socket.recv(5))
        payload = b""

        for _ in range(int(data_size / 4096)):
            payload += self.socket.recv(4096)

        payload += self.socket.recv(int(data_size % 4096))

        try:
            if payload[0] == ETH_ANSWER_ERR:
                raise TimeoutError("Server timed out waiting for serial response")
        except IndexError:
            self._reset_input_buffer()
            raise SerialErrPckgLen(
                "Received empty response, check if the controller is on and connected. If you receive garbled output, try disconnecting and reconnecting."
            )

        return payload

    @validate
    def _transfer(self, msg: str, size: int) -> bytes:
        base_msg = (self._slave_addr + msg).encode("ISO-8859-1")
        full_msg = self._format_message(checksum(base_msg), ETH_CMD_REQUEST)
        self.socket.sendall(full_msg)
        return self._get_reply(size)

    def _transfer_write(self, msg: str):
        base_msg = (self._slave_addr + msg).encode("ISO-8859-1")
        full_msg = self._format_message(checksum(base_msg), ETH_CMD_REQUEST)
        self.socket.sendall(full_msg)
        try:
            self._get_reply()
            return
        except:
            return

    @property
    def timeout(self) -> float:
        return self.socket.timeout

    @timeout.setter
    def timeout(self, new_timeout: float):
        self._serial_timeout = new_timeout * 1000
        self.socket.settimeout(new_timeout)

    def is_open(self) -> bool:
        try:
            data = self.socket.recv(4, socket.MSG_DONTWAIT | socket.MSG_PEEK)
            if len(data) == 0:
                return False
        except BlockingIOError:
            return True
        except ConnectionResetError:
            return False
        return True

    def connect(self, address: str = "127.0.0.1", port: int = 5000):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.settimeout(2)

        self.socket.connect((address, port))
        self.socket.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)

    def disconnect(self):
        self.socket.close()


def GenericDRS(com_or_address: str, port_or_baud: int):
    """Factory for DRS communication classes"""
    if re.match(r"(([0-9]{1,3}\.){3}[0-9]{1,3})", com_or_address):
        return EthDRS(com_or_address, port_or_baud)
    else:
        return SerialDRS(com_or_address, port_or_baud)



from warnings import warn
from .consts import ETH_ANSWER_NOQUEUE

from .utils import checksum


class SerialError(Exception):
    """Exception raised when there's a problem with serial."""


class SerialErrCheckSum(SerialError):
    """Exception raised when there'a problem with checksum."""


class SerialErrPckgLen(SerialError):
    """Exception raised when there'a problem with package length."""


class SerialForbidden(SerialError):
    """Exception raised when a given operation is not permitted by the UDC"""


class SerialInvalidCmd(SerialError):
    """Exception raised when the supplied command is invalid"""


SERIAL_ERROR = [
    "Ok",
    "Power supply in local mode",
    "Power supply in PC host mode",
    "Power supply interlocked",
    "UDC stuck",
    "DSP timeout",
    "DSP busy",
    "Resource is busy",
    "Invalid command",
]

ERROR_RESPONSE = {
    0xE1: "Malformed message",
    0xE2: "Operation not supported",
    0xE3: "Invalid ID",
    0xE4: "Invalid value",
    0xE5: "Invalid payload size",
    0xE6: "Read-only",
    0xE7: "Insufficient memory",
    0xE8: "Resource busy",
}


def validate(func):
    def wrapper(*args, **kwargs):
        reply = func(*args, **kwargs)
        if len(reply) == 0 or (len(reply) == 1 and reply[0] == ETH_ANSWER_NOQUEUE):
            args[0]._reset_input_buffer()
            raise SerialErrPckgLen(
                "Received empty response, check if the controller is on and connected. If you receive garbled output, try disconnecting and reconnecting."
            )

        reply = reply[1:] if len(reply) - 1 == args[2] else reply
        check_serial_error(reply)

        if len(reply) != args[2] or (len(reply) - 1 != args[2] and reply[0] == 0x21):
            offset = 1 if reply[0] == 0x21 else 0
            if len(reply) > 5:
                check_serial_error(reply[offset:])

            args[0]._reset_input_buffer()
            raise SerialErrPckgLen(
                "Expected {} bytes, received {} bytes".format(
                    args[2], len(reply) - offset
                )
            )

        if reply != checksum(reply[:-1]):
            args[0]._reset_input_buffer()
            raise SerialErrCheckSum(
                "Expected {} as checksum, received {}".format(
                    checksum(reply[:-1])[-1], reply[-1]
                )
            )

        return reply

    return wrapper


def print_deprecated(func):
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def check_serial_error(reply: bytes):
    error_index = 2 if reply[0] == 0x21 else 1
    if reply[error_index] == 0x53:
        if reply[-2] == 4:
            raise SerialForbidden(
                "Command blocked by UDC, unlock with unlock_udc() first"
            )
        if reply[-2] == 8:
            raise SerialInvalidCmd
        else:
            raise SerialError(SERIAL_ERROR[reply[-2]])
    elif reply[error_index] in ERROR_RESPONSE:
        raise SerialInvalidCmd(ERROR_RESPONSE[reply[error_index]])
